import itertools
import json
import logging
import time
from pathlib import Path
from typing import Any, ClassVar, List

import pandas as pd
import pyspark.sql.functions as F
from pyspark.ml import Transformer


class UserDefinedFunction:
    """
    User defined function in Python.

    .. versionadded:: 1.3

    Notes
    -----
    The constructor of this class is not supposed to be directly called.
    Use :meth:`pyspark.sql.functions.udf` or :meth:`pyspark.sql.functions.pandas_udf`
    to create this instance.

    """

    def __init__(
        self,
        func,
        returnType,
        name,
        evalType,
        deterministic: bool = True,
    ) -> None:
        if not callable(func):
            raise ValueError("Invalid function: not a function or callable")
        self.func = func
        self.returnType = returnType

    def __call__(self, *cols: Any) -> Any:
        cols = zip(*cols)
        return [self.func(*i) for i in cols]

    def _wrapped(self) -> Any:
        return self


def lit(value) -> Any:
    """Creates a :class:`Column` of literal value."""
    return itertools.repeat(value)


def explode(col: Any) -> Any:
    return col


def _invoke_function(name: str, *args: Any) -> Any:
    if name == "lit":
        return lit(*args)
    raise ValueError("Invalid function name: %s" % name)


temp_functions = {}


def pathSparkFunctions(pyspark: Any) -> None:
    """Path Spark functions."""
    temp_functions["udf"] = pyspark.sql.udf.UserDefinedFunction
    pyspark.sql.udf.UserDefinedFunction = UserDefinedFunction
    temp_functions["invoke_function"] = pyspark.sql.functions._invoke_function
    pyspark.sql.functions._invoke_function = _invoke_function


def unpathSparkFunctions(pyspark: Any) -> None:
    """Unpath Spark functions."""
    pyspark.sql.udf.UserDefinedFunction = temp_functions["udf"]
    pyspark.sql.functions._invoke_function = temp_functions["invoke_function"]


def posexplode(df, in_col, pos_col="page", val_col="value"):
    """Explode a column of lists into multiple rows, with position index."""

    if isinstance(df, pd.DataFrame):

        def to_list(x):
            val = x[in_col]
            if not isinstance(val, list):
                try:
                    val = list(val)  # will work for generator, tuple, etc.
                except TypeError:
                    val = [val]
            return DatasetPd({pos_col: range(len(val)), val_col: val})

        out = df[[in_col]].apply(to_list, axis=1)
        out = pd.concat(out.tolist(), keys=df.index).reset_index()
        return DatasetPd(
            df.drop(in_col).merge(out, left_index=True, right_on="level_0"),
        )
    else:
        sel_col = [
            *df.columns,
            *[
                F.posexplode(
                    F.col(in_col),
                ).alias(pos_col, val_col),
            ],
        ]

        return df.select(*sel_col)


class DatasetPd(pd.DataFrame):

    def withColumn(self, name, col) -> "DatasetPd":
        if name in self.columns:
            self[name] = col
        else:
            self.insert(0, name, col, True)
        return self

    def drop(self, col) -> "DatasetPd":
        return DatasetPd(super().drop(columns=[col]))

    def repartition(self, numPartitions) -> "DatasetPd":
        return self

    def coalesce(self, numPartitions) -> "DatasetPd":
        return self

    def select(self, *args):
        return self[list(args)]


class PandasPipeline:
    """PandasPipeline for call Spark ML Pipelines with Pandas DataFrame."""

    stages: ClassVar[List[Transformer]] = []

    def setStages(self, value) -> "PandasPipeline":
        self.stages = value
        return self

    def __init__(self, stages) -> None:
        self.setStages(stages)

    def fromFile(self, filename: str) -> Any:
        with Path.open(filename, "rb") as f:
            data = f.read()

        data = DatasetPd({"content": [data], "path": [filename], "resolution": [0]})

        return self.fromPandas(data)

    def fromBinary(self, data, filename="memory") -> Any:

        data = DatasetPd({"content": [data], "path": [filename], "resolution": [0]})
        return self.fromPandas(data)

    def fromPandas(self, data: pd.DataFrame) -> Any:

        start_time_total = time.time()
        execution_times = {"stages": {}}
        data = DatasetPd(data)
        for stage in self.stages:

            stage_name = stage.__class__.__name__
            start_time_stage = time.time()
            data = stage._transform(data)

            stage_duration = time.time() - start_time_stage
            execution_times["stages"][stage_name] = stage_duration
            logging.info(
                f"Stage {stage_name} completed in {stage_duration:.2f} seconds",
            )

        total_duration = time.time() - start_time_total
        execution_times["total"] = total_duration
        logging.info(f"Total execution time: {total_duration:.2f} seconds")

        # Add execution time information as a new column
        data["execution_time"] = json.dumps(execution_times)

        return data

    def fromDict(self, data: dict) -> Any:
        data = DatasetPd(data)
        return self.fromPandas(data)



from types import MappingProxyType
from typing import Any

import numpy
import torch
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice, Param, Params, TypeConverters
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class CraftTextDetector(BaseDetector, HasDevice, HasBatchSize):
    _craft_net = None
    _refine_net = None

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.7,
            "textThreshold": 0.4,
            "linkThreshold": 0.4,
            "sizeThreshold": -1,
            "width": 1280,
            "withRefiner": False,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
        },
    )

    textThreshold = Param(
        Params._dummy(),
        "textThreshold",
        "Threshold for text region score",
        typeConverter=TypeConverters.toFloat,
    )

    linkThreshold = Param(
        Params._dummy(),
        "linkThreshold",
        "Threshold for link affinity score",
        typeConverter=TypeConverters.toFloat,
    )

    sizeThreshold = Param(
        Params._dummy(),
        "sizeThreshold",
        "Threshold for height of detected regions",
        typeConverter=TypeConverters.toInt,
    )

    width = Param(
        Params._dummy(),
        "width",
        "Width for image resizing",
        typeConverter=TypeConverters.toInt,
    )

    withRefiner = Param(
        Params._dummy(),
        "withRefiner",
        "Enable refiner network postprocessing",
        typeConverter=TypeConverters.toBoolean,
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(CraftTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._craft_net and cls._refine_net:
            return cls._craft_net, cls._refine_net

        from craft_text_detector import load_craftnet_model, load_refinenet_model

        device = "cuda" if int(params["device"]) == Device.CUDA.value else "cpu"
        use_cuda = device == "cuda"

        craft_net = load_craftnet_model(
            cuda=use_cuda,
        )

        refine_net = None
        if params.get("withRefiner"):
            refine_net = load_refinenet_model(
                cuda=use_cuda,
            )

        cls._craft_net = craft_net
        cls._refine_net = refine_net
        return craft_net, refine_net

    @classmethod
    def call_detector(cls, images, params):
        import cv2
        from craft_text_detector import craft_utils, image_utils, torch_utils

        craft_net, refine_net = cls.get_model(params)
        use_cuda = int(params["device"]) == Device.CUDA.value
        results = []

        for image, image_path in images:
            try:
                # Convert PIL to OpenCV format
                #img_cv = numpy.array(image)[:, :, ::-1]

                img_cv = image_utils.read_image(numpy.array(image)[:, :, ::-1])

                # Resize
                img_resized, target_ratio, size_heatmap = image_utils.resize_aspect_ratio(
                    img_cv,
                    params["width"],
                    interpolation=cv2.INTER_LINEAR,
                )
                ratio_h = ratio_w = 1 / target_ratio


                # Preprocess
                x = image_utils.normalizeMeanVariance(img_resized)
                x = torch_utils.from_numpy(x).permute(2, 0, 1)
                x = torch_utils.Variable(x.unsqueeze(0))
                if use_cuda:
                    x = x.cuda()

                # Forward pass
                with torch_utils.no_grad():
                    y, feature = craft_net(x)

                score_text = y[0, :, :, 0].cpu().data.numpy()
                score_link = y[0, :, :, 1].cpu().data.numpy()

                # Refine if enabled
                if refine_net is not None:
                    with torch_utils.no_grad():
                        y_refiner = refine_net(y, feature)
                    score_link = y_refiner[0, :, :, 0].cpu().data.numpy()

                # Post-process
                boxes, _ = craft_utils.getDetBoxes(
                    score_text,
                    score_link,
                    params["scoreThreshold"],
                    params["linkThreshold"],
                    params["textThreshold"],
                    poly=False,
                )

                # Adjust coordinates
                boxes = craft_utils.adjustResultCoordinates(boxes, ratio_w, ratio_h)


                # Convert to Box objects
                box_objects = [Box.from_polygon(box) for box in boxes]

                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=box_objects,
                        exception="",
                    ),
                )

            except Exception as e:
                raise e
                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=[],
                        exception=f"CraftTextDetector error: {e!s}",
                    ),
                )

        if use_cuda:
            torch.cuda.empty_cache()

        return results

    def setTextThreshold(self, value):
        return self._set(textThreshold=value)

    def setLinkThreshold(self, value):
        return self._set(linkThreshold=value)

    def setSizeThreshold(self, value):
        return self._set(sizeThreshold=value)

    def setWidth(self, value):
        return self._set(width=value)

    def setWithRefiner(self, value):
        return self._set(withRefiner=value)



# from pyspark.ml.param.shared import HasInputCol, HasOutputCol
# from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable, JavaMLReadable, JavaMLWritable
# from pyspark.ml.wrapper import JavaTransformer
# from pyspark.ml.param import Param, Params, TypeConverters
# from pyspark.sql.types import *
# from pyspark import keyword_only
# from pathlib import Path
#
# import json
# import time
# import logging
# import traceback
#
# import pyspark.sql.functions as f
# import torch
# from pyspark.ml import Transformer
# from cleverdoc.schemas import Box, Image
# from cleverdoc.params import HasKeepInput
#
#
#
# class TextDetector(Transformer, HasInputCol, HasOutputCol,
#                           HasKeepInput, DefaultParamsReadable, DefaultParamsWritable, JavaMLReadable, JavaMLWritable):
#     name = "TextDetector"
#
#     scoreThreshold = Param(Params._dummy(), "scoreThreshold",
#                            "Score threshold above which detection is considered as reliable",
#                            typeConverter=TypeConverters.toFloat)
#     textThreshold = Param(Params._dummy(), "textThreshold",
#                           "Threshold for the region(text) score. The region score represents \
#                           the probability that the given pixel is the center of the character",
#                           typeConverter=TypeConverters.toFloat)
#     linkThreshold = Param(Params._dummy(), "linkThreshold",
#                           "Threshold for the the link(affinity) score. The link score represents \
#                           the center probability of the space between adjacent characters",
#                           typeConverter=TypeConverters.toFloat)
#     sizeThreshold = Param(Params._dummy(), "sizeThreshold",
#                           "Threshold for height of the detected regions",
#                           typeConverter=TypeConverters.toInt)
#     withRefiner = Param(Params._dummy(), "withRefiner",
#                         "Enable to run refiner net as postprocessing step.",
#                         typeConverter=TypeConverters.toBoolean)
#     width = Param(Params._dummy(), "width",
#                   "Width of the desired input image. Image will be resized to this width.",
#                   typeConverter=TypeConverters.toInt)
#
#     mergeIntersects = Param(Params._dummy(), "mergeIntersects",
#                             "Merge intersects boxes.",
#                             typeConverter=TypeConverters.toBoolean)
#
#     useGPU = Param(Params._dummy(), "useGPU",
#                    "Try to use the GPU if available.",
#                    typeConverter=TypeConverters.toBoolean)
#
#     usePandasUdf = Param(Params._dummy(), "usePandasUdf",
#                          "Use Pandas Udf.",
#                          typeConverter=TypeConverters.toBoolean)
#
#     @keyword_only
#     def __init__(self):
#         super(TextDetector, self).__init__()
#
#         self._setDefault(scoreThreshold=0.7)
#         self._setDefault(textThreshold=0.4)
#         self._setDefault(linkThreshold=0.4)
#         self._setDefault(width=1280)
#         self._setDefault(sizeThreshold=-1)
#         self._setDefault(mergeIntersects=True)
#         self._setDefault(forceProcessing=True)
#         self._setDefault(keepInput=True)
#         self._setDefault(useGPU=False)
#         self._setDefault(usePandasUdf=False)
#
#
#     def setInputCol(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(inputCol=value)
#
#     def setOutputCol(self, value):
#         """
#         Sets the value of :py:attr:`outputCol`.
#         """
#         return self._set(outputCol=value)
#
#     @staticmethod
#     def get_prediction(
#             image,
#             craft_net,
#             refine_net=None,
#             text_threshold: float = 0.7,
#             link_threshold: float = 0.4,
#             low_text: float = 0.4,
#             cuda: bool = False,
#             long_size: int = 1280,
#             poly: bool = True,
#     ):
#
#         import cv2
#
#         from craft_text_detector import craft_utils
#         from craft_text_detector import image_utils
#         from craft_text_detector import torch_utils
#         # pylint: disable=pointless-string-statement
#         """
#         Arguments:scoreThreshold
#             image: path to the image to be processed or numpy array or PIL image
#             output_dir: path to the results to be exported
#             craft_net: craft net model
#             refine_net: refine net model
#             text_threshold: text confidence threshold
#             link_threshold: link confidence threshold
#             low_text: text low-bound score
#             cuda: Use cuda for inference
#             canvas_size: image size for inference
#             long_size: desired longest image size for inference
#             poly: enable polygon type
#         Output:
#             {"masks": lists of predicted masks 2d as bool array,
#              "boxes": list of coords of points of predicted boxes,
#              "boxes_as_ratios": list of coords of points of predicted boxes as ratios of image size,
#              "polys_as_ratios": list of coords of points of predicted polys as ratios of image size,
#              "heatmaps": visualizations of the detected characters/links,
#              "times": elapsed times of the sub modules, in seconds}
#         """
#         t0 = time.time()
#
#         # read/convert image
#         image = image_utils.read_image(image)
#
#         # resize
#         img_resized, target_ratio, size_heatmap = image_utils.resize_aspect_ratio(
#             image, long_size, interpolation=cv2.INTER_LINEAR
#         )
#         ratio_h = ratio_w = 1 / target_ratio
#         resize_time = time.time() - t0
#         t0 = time.time()
#
#         # preprocessing
#         x = image_utils.normalizeMeanVariance(img_resized)
#         x = torch_utils.from_numpy(x).permute(2, 0, 1)  # [h, w, c] to [c, h, w]
#         x = torch_utils.Variable(x.unsqueeze(0))  # [c, h, w] to [b, c, h, w]
#         if cuda:
#             x = x.cuda()
#         preprocessing_time = time.time() - t0
#         t0 = time.time()
#
#         # forward pass
#         with torch_utils.no_grad():
#             y, feature = craft_net(x)
#         craftnet_time = time.time() - t0
#         t0 = time.time()
#
#         # make score and link map
#         score_text = y[0, :, :, 0].cpu().data.numpy()
#         score_link = y[0, :, :, 1].cpu().data.numpy()
#
#         # refine link
#         if refine_net is not None:
#             with torch_utils.no_grad():
#                 y_refiner = refine_net(y, feature)
#             score_link = y_refiner[0, :, :, 0].cpu().data.numpy()
#         refinenet_time = time.time() - t0
#         t0 = time.time()
#
#         # Post-processing
#         boxes, polys = craft_utils.getDetBoxes(
#             score_text, score_link, text_threshold, link_threshold, low_text, poly
#         )
#
#         # coordinate adjustment
#         boxes = craft_utils.adjustResultCoordinates(boxes, ratio_w, ratio_h)
#
#         postprocess_time = time.time() - t0
#
#         times = {
#             "resize_time": resize_time,
#             "preprocessing_time": preprocessing_time,
#             "craftnet_time": craftnet_time,
#             "refinenet_time": refinenet_time,
#             "postprocess_time": postprocess_time,
#         }
#
#         return {
#             "boxes": boxes,
#             "times": times,
#         }
#
#     @staticmethod
#     def udf(image, model_path, params, exception):
#         logging.info("Run ImageTextDetectorV2")
#
#         try:
#
#             params = json.loads(params)
#             use_cuda = torch.cuda.is_available() and params.get("useGPU")
#
#             from craft_text_detector import (
#                 load_craftnet_model,
#                 load_refinenet_model,
#                 # get_prediction,
#                 empty_cuda_cache
#             )
#             # read image
#             image = to_opencv_image(image)
#
#             # load models
#             refine_net = None
#             if params.get("withRefiner"):
#                 refine_net = load_refinenet_model(cuda=use_cuda,
#                                                   weight_path=str(Path(model_path, "craft_refiner_CTW1500.pth")))
#             craft_net = load_craftnet_model(cuda=use_cuda, weight_path=str(Path(model_path, "craft_mlt_25k.pth")))
#
#             long_size = int(params.get("width") if params.get("width") != 0 else 1280)
#             # perform prediction
#             prediction_result = ImageTextDetectorV2.get_prediction(
#                 image=image,
#                 craft_net=craft_net,
#                 refine_net=refine_net,
#                 text_threshold=params.get("scoreThreshold"),
#                 link_threshold=params.get("linkThreshold"),
#                 low_text=params.get("textThreshold"),
#                 cuda=use_cuda,
#                 long_size=long_size,
#                 poly=False
#             )
#
#             rects = [Rectangle.from_box(box) for box in prediction_result["boxes"]]
#             character_height = get_size(rects, lambda x: x.get_height())
#
#             size_threshold = params.get("sizeThreshold")
#             if params.get("sizeThreshold") == -1:
#                 size_threshold = int(character_height / 2)
#             line_tolerance = int(character_height / 2)
#
#             rects = list(filter(lambda x: x.get_height() > size_threshold, rects))
#
#             if params.get("mergeIntersects"):
#                 rects = Rectangle.merge_intersects(rects, line_tolerance)
#
#             regions = [rect.to_region() for rect in rects]
#         except Exception as error:
#             logging.exception(error)
#             regions = []
#             exception = "ImageTextDetectorV2: Error during text detection: " + str(traceback.format_exc()) + str(
#                 exception)
#         finally:
#             if use_cuda:
#                 empty_cuda_cache()
#
#         return (regions, exception)
#
#     @staticmethod
#     def pandas_udf(images, model_path, params, exception):
#         logging.info("Run ImageTextDetectorV2")
#
#         params = json.loads(params[0])
#         use_cuda = torch.cuda.is_available() and params.get("useGPU")
#
#         exception = exception[0]
#
#         import pandas as pd
#
#         from craft_text_detector import (
#             load_craftnet_model,
#             load_refinenet_model,
#             empty_cuda_cache
#         )
#
#         # load models
#         refine_net = None
#         if params.get("withRefiner"):
#             refine_net = load_refinenet_model(cuda=use_cuda,
#                                               weight_path=str(Path(model_path[0], "craft_refiner_CTW1500.pth")))
#         craft_net = load_craftnet_model(cuda=use_cuda, weight_path=str(Path(model_path[0], "craft_mlt_25k.pth")))
#
#         long_size = int(params.get("width") if params.get("width") != 0 else 1280)
#
#         res = []
#         for i, image in images.iterrows():
#             logging.info(f"ImageTextDetectorV2: process image {i}")
#             try:
#                 # read image
#                 image = to_opencv_image(Image(**image))
#                 prediction_result = ImageTextDetectorV2.get_prediction(
#                     image=image,
#                     craft_net=craft_net,
#                     refine_net=refine_net,
#                     text_threshold=params.get("scoreThreshold"),
#                     link_threshold=params.get("linkThreshold"),
#                     low_text=params.get("textThreshold"),
#                     cuda=use_cuda,
#                     long_size=long_size,
#                     poly=False
#                 )
#
#                 rects = [Rectangle.from_box(box) for box in prediction_result["boxes"]]
#                 character_height = get_size(rects, lambda x: x.get_height())
#
#                 size_threshold = params.get("sizeThreshold")
#                 if params.get("sizeThreshold") == -1:
#                     size_threshold = int(character_height / 2)
#                 line_tolerance = int(character_height / 2)
#
#                 rects = list(filter(lambda x: x.get_height() > size_threshold, rects))
#
#                 if params.get("mergeIntersects"):
#                     rects = Rectangle.merge_intersects(rects, line_tolerance)
#
#                 regions = [rect.to_region() for rect in rects]
#
#             except Exception as error:
#                 logging.exception(error)
#                 regions = []
#                 exception = "ImageToTextV2: Error during text detection: " + traceback.format_exc() + exception
#             res.append(json.dumps({params['outputCol']: regions, "exception": exception}))
#         if use_cuda:
#             empty_cuda_cache()
#
#         return pd.Series(res)
#
#     def _transform(self, dataset):
#
#         out_col = self.getOutputCol()
#
#         if self.getInputCol() not in dataset.columns:
#             uid_ = self.uid
#             inp_col = self.getInputCol()
#             raise ValueError(f"""Missing input column in {uid_}: Column '{inp_col}' is not present.
#         Make sure such transformer exist in your pipeline,
#         with the right output names.""")
#
#         image_col = self.getInputCol()
#
#         params = json.dumps({k.name: v for k, v in self.extractParamMap().items()})
#
#         output_schema = StructType([StructField(out_col, CoordinateSchema, True),
#                                     StructField('exception', StringType(), True)])
#
#         if "exception" not in dataset.columns:
#             dataset = dataset.withColumn("exception", f.lit(""))
#
#         if self.getOrDefault(self.usePandasUdf):
#             sel_cols_input = [f.from_json(f.pandas_udf(ImageTextDetectorV2.pandas_udf, StringType())(image_col,
#                                                                                                      f.lit(
#                                                                                                          self.getModelPath()),
#                                                                                                      f.lit(params),
#                                                                                                      "exception"),
#                                           output_schema).alias("tmp_result")] + \
#                              [f.col(x) for x in dataset.columns if
#                               x not in (image_col, "exception")]
#         else:
#             sel_cols_input = [self.exploding_wrap(f.udf(ImageTextDetectorV2.udf, output_schema))(image_col,
#                                                                                                  f.lit(
#                                                                                                      self.getModelPath()),
#                                                                                                  f.lit(params),
#                                                                                                  "exception").alias(
#                 "tmp_result")] + \
#                              [f.col(x) for x in dataset.columns if
#                               x not in (image_col, "exception")]
#
#         sel_cols = [f.col("tmp_result.*")] + \
#                    [f.col(x) for x in dataset.columns if x not in (image_col, "exception")]
#
#         if self.getOrDefault(self.keepInput):
#             sel_cols_input.append(f.col(image_col))
#             sel_cols.append(f.col(image_col))
#
#         result = dataset.select(*sel_cols_input) \
#             .select(*sel_cols)
#         if not self.getOrDefault(self.keepInput):
#             result = result.drop(image_col)
#         return result
#
#     def setTextThreshold(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(textThreshold=value)
#
#     def setLinkThreshold(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(linkThreshold=value)
#
#     def setScoreThreshold(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(scoreThreshold=value)
#
#     def setWithRefiner(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(withRefiner=value)
#
#     def setUseGPU(self, value):
#         """
#         Sets the value of :py:attr:`useGPU`.
#         """
#         return self._set(useGPU=value)
#
#     def setSizeThreshold(self, value):
#         """
#         Sets the value of :py:attr:`sizeThreshold`.
#         """
#         return self._set(sizeThreshold=value)
#
#     def setWidth(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(width=value)
#
#     def setMergeIntersects(self, value):
#         """
#         Sets the value of :py:attr:`mergeIntersects`.
#         """
#         return self._set(mergeIntersects=value)
#
#     def setUsePandasUdf(self, value):
#         """
#         Sets the value of :py:attr:`spaceWidth`.
#         """
#         return self._set(usePandasUdf=value)



import logging
import traceback
from types import MappingProxyType
from typing import Any

import fitz
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.pandas import DataFrame
from pyspark.sql.functions import udf
from pyspark.sql.types import ArrayType, Row

from scaledp.enums import ImageType
from scaledp.params import (
    HasColumnValidator,
    HasDefaultEnum,
    HasImageType,
    HasInputCol,
    HasKeepInputData,
    HasOutputCol,
    HasPageCol,
    HasPathCol,
    HasResolution,
    Param,
    Params,
    TypeConverters,
)
from scaledp.pipeline.PandasPipeline import posexplode
from scaledp.schemas.Image import Image


class PdfDataToImage(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasPathCol,
    HasResolution,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColumnValidator,
    HasDefaultEnum,
):
    """Extract image from PDF file."""

    pageLimit = Param(
        Params._dummy(),
        "pageLimit",
        "Limit number of pages to convert to image",
        typeConverter=TypeConverters.toInt,
    )

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCol": "content",
            "outputCol": "image",
            "pathCol": "path",
            "pageCol": "page",
            "keepInputData": False,
            "imageType": ImageType.FILE,
            "resolution": 300,
            "pageLimit": 0,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(PdfDataToImage, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def transform_udf(self, input: Row, path: Row) -> list[Image]:
        logging.info("Run Pdf Data to Image")
        try:
            doc = fitz.open("pdf", input)
            if len(doc) == 0:
                raise ValueError("Empty PDF document.")
            if self.getPageLimit():
                doc = doc[: self.getPageLimit()]
            for page in doc:
                pix = page.get_pixmap(
                    matrix=fitz.Identity,
                    dpi=self.getResolution(),
                    colorspace=fitz.csRGB,
                    clip=None,
                    alpha=False,
                    annots=True,
                )
                yield Image.from_binary(
                    pix.pil_tobytes("png"),
                    path,
                    self.getImageType(),
                    resolution=self.getResolution(),
                    width=pix.width,
                    height=pix.height,
                )

        except Exception:
            exception = traceback.format_exc()
            exception = (
                f"{self.uid}: Error during extract image from "
                f"the PDF document: {exception}"
            )
            logging.warning(exception)
            return [Image(path=path, exception=exception)]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        path_col = dataset[self.getPathCol()]

        df_1 = dataset.withColumn(
            "temp_data",
            udf(self.transform_udf, ArrayType(Image.get_schema()))(
                input_col,
                path_col,
            ),
        )

        result = posexplode(df_1, "temp_data", self.getPageCol(), out_col)

        if not self.getKeepInputData():
            result = result.drop(self.getInputCol())
        return result

    def getPageLimit(self) -> int:
        """Get page limit."""
        return self.getOrDefault(self.pageLimit)

    def setPageLimit(self, value) -> "PdfDataToImage":
        """Set page limit."""
        return self._set(pageNumber=value)



import gc
from types import MappingProxyType
from typing import Any

import numpy as np
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.models.recognizers.BaseOcr import BaseOcr
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document


class EasyOcr(BaseOcr, HasDevice, HasBatchSize):

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "text",
            "bypassCol": "",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "lang": ["eng"],
            "lineTolerance": 0,
            "keepFormatting": False,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "device": Device.CPU,
            "batchSize": 2,
            "propagateError": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(EasyOcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @staticmethod
    def points_to_box(points, text, score):
        """Convert a set of four corner points to (x, y, width, height)."""
        x_coords = [p[0] for p in points]
        y_coords = [p[1] for p in points]

        x = min(x_coords)
        y = min(y_coords)
        width = max(x_coords) - x
        height = max(y_coords) - y

        return Box(text=text, score=score, x=x, y=y, width=width, height=height)

    @classmethod
    def call_ocr(cls, images, params):
        import easyocr
        import torch

        device = int(params["device"]) != Device.CPU.value

        langs = params["lang"]
        scale_factor = params["scaleFactor"]
        reader = easyocr.Reader(langs, device)
        results = []
        for img, image_path in images:
            image = np.array(img.convert("RGB"))[:, :, ::-1].copy()
            result = reader.readtext(image)
            boxes = [
                EasyOcr.points_to_box(box, text, float(score))
                .to_string()
                .scale(1 / scale_factor)
                for box, text, score in result
            ]

            if params["keepFormatting"]:
                text = EasyOcr.box_to_formatted_text(boxes, params["lineTolerance"])
            else:
                text = "\n".join([str(w.text) for w in boxes])

            results.append(
                Document(path=image_path, text=text, type="text", bboxes=boxes),
            )

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results



import os
import sys
from importlib import metadata, resources
from importlib.util import find_spec

import pyspark
from pyspark.ml.pipeline import PipelineModel
from pyspark.sql import DataFrame, SparkSession

from scaledp import enums
from scaledp.enums import *  # noqa
from scaledp.image.DataToImage import DataToImage
from scaledp.image.ImageCropBoxes import ImageCropBoxes
from scaledp.image.ImageDrawBoxes import ImageDrawBoxes
from scaledp.models.detectors.DocTRTextDetector import DocTRTextDetector
from scaledp.models.detectors.YoloDetector import YoloDetector
from scaledp.models.extractors.DSPyExtractor import DSPyExtractor
from scaledp.models.extractors.LLMExtractor import LLMExtractor
from scaledp.models.extractors.LLMVisualExtractor import LLMVisualExtractor
from scaledp.models.ner.LLMNer import LLMNer
from scaledp.models.ner.Ner import Ner
from scaledp.models.recognizers.DocTROcr import DocTROcr
from scaledp.models.recognizers.EasyOcr import EasyOcr
from scaledp.models.recognizers.LLMOcr import LLMOcr
from scaledp.models.recognizers.SuryaOcr import SuryaOcr
from scaledp.models.recognizers.TesseractOcr import TesseractOcr
from scaledp.models.recognizers.TesseractRecognizer import TesseractRecognizer
from scaledp.pdf.PdfDataToImage import PdfDataToImage
from scaledp.text.TextToDocument import TextToDocument
from scaledp.utils.show_utils import (
    show_image,
    show_json,
    show_ner,
    show_pdf,
    show_text,
    visualize_ner,
)

DataFrame.show_image = (
    lambda self, column="", limit=5, width=None, show_meta=True: show_image(
        self,
        column,
        limit,
        width,
        show_meta,
    )
)
DataFrame.show_pdf = (
    lambda self, column="", limit=5, width=None, show_meta=True: show_pdf(
        self,
        column,
        limit,
        width,
        show_meta,
    )
)
DataFrame.show_ner = lambda self, column="ner", limit=20, truncate=True: show_ner(
    self,
    column,
    limit,
    truncate,
)
DataFrame.show_text = (
    lambda self, column="", field="text", limit=20, width=None: show_text(
        self,
        column,
        field,
        limit,
        width,
    )
)
DataFrame.show_json = (
    lambda self, column="", field="json_data", limit=20, width=None: show_json(
        self,
        column,
        field,
        limit,
        width,
    )
)
DataFrame.visualize_ner = (
    lambda self, column="ner", text_column="text", limit=20, width=None: visualize_ner(
        self,
        column,
        text_column,
        limit,
        width,
    )
)


def version():
    return metadata.version("scaledp")


__version__ = version()


def files(path):
    """File resources."""
    return resources.files("scaledp").joinpath(path).as_posix()


SPARK_PDF_VERSION = "0.1.15"


def aws_version():
    spark_hadoop_map = {
        "3.0": "2.7.4",
        "3.1": "3.2.0",
        "3.2": "3.3.1",
        "3.3": "3.3.2",
        "3.4": "3.3.4",
        "3.5": "3.3.4",
    }
    return spark_hadoop_map[pyspark.__version__[:3]]


def spark_version() -> str:
    return pyspark.__version__[:3].replace(".", "")


def scala_version() -> str:
    if int(spark_version()) >= 40:
        return "2.13"
    return "2.12"


def ScaleDPSession(
    conf=None,
    master_url="local[*]",
    with_aws=False,
    with_pro=False,
    with_spark_pdf: str | bool = True,
    logLevel="ERROR",
):
    """
    Start Spark session with ScaleDP.

    @param conf: Instance of SparkConf or dict with extra configuration.
    @param master_url: Spark master URL
    @param with_aws: Enable AWS support
    @param logLevel: Log level
    """
    os.environ["PYSPARK_PYTHON"] = sys.executable
    os.environ["TRANSFORMERS_VERBOSITY"] = logLevel.lower()
    os.environ["PYARROW_IGNORE_TIMEZONE"] = "1"

    if with_pro and find_spec("scaledp_pro") is None:
        raise ImportError(
            "ScaleDP Pro is not installed. Please install it using 'pip install scaledp-pro'",
        )

    jars = []
    jars_packages = []
    default_conf = {
        "spark.serializer": "org.apache.spark.serializer.KryoSerializer",
        "spark.kryoserializer.buffer.max": "200M",
        "spark.driver.memory": "8G",
    }

    if with_aws:
        jars_packages.append("org.apache.hadoop:hadoop-aws:" + aws_version())

    if with_spark_pdf:
        spark_pdf_version = SPARK_PDF_VERSION
        if isinstance(with_spark_pdf, str):
            spark_pdf_version = with_spark_pdf
        jars_packages.append(
            f"com.stabrise:spark-pdf-spark{spark_version()}_{scala_version()}:{spark_pdf_version}",
        )

    if conf:
        if not isinstance(conf, dict):
            conf = dict(conf.getAll())
        default_conf.update(conf)
        extra_jars_packages = default_conf.get("spark.jars.packages")
        if extra_jars_packages:
            jars_packages.append(extra_jars_packages)
        extra_jars = default_conf.get("spark.jars")
        if extra_jars:
            jars.append(extra_jars)

    builder = SparkSession.builder.master(master_url).appName(
        f"ScaleDP: v{__version__}"
        + (f" Spark PDF: v{spark_pdf_version}" if with_spark_pdf else ""),
    )

    for k, v in default_conf.items():
        builder.config(str(k), str(v))

    builder.config("spark.jars", ",".join(jars))
    builder.config("spark.jars.packages", ",".join(jars_packages))

    spark = builder.getOrCreate()
    spark.sparkContext.setLogLevel(logLevel=logLevel)
    return spark


__all__ = [
    "ScaleDPSession",
    "DataToImage",
    "ImageDrawBoxes",
    "PdfDataToImage",
    "TesseractOcr",
    "Ner",
    "TextToDocument",
    "PipelineModel",
    "SuryaOcr",
    "EasyOcr",
    "DocTROcr",
    "YoloDetector",
    "ImageCropBoxes",
    "DSPyExtractor",
    "TesseractRecognizer",
    "DocTRTextDetector",
    "LLMVisualExtractor",
    "LLMExtractor",
    "LLMOcr",
    "LLMNer",
    "__version__",
    "files",
    *dir(enums),
]



LANGUAGE_TO_TESSERACT_CODE = {
    "Afrikaans": "afr",
    "Amharic": "amh",
    "Arabic": "ara",
    "Assamese": "asm",
    "Azerbaijani": "aze",
    "Belarusian": "bel",
    "Bulgarian": "bul",
    "Bengali": "ben",
    "Breton": "bre",
    "Bosnian": "bos",
    "Catalan": "cat",
    "Czech": "ces",
    "Welsh": "cym",
    "Danish": "dan",
    "German": "deu",
    "Greek": "ell",
    "English": "eng",
    "Esperanto": "epo",
    "Spanish": "spa",
    "Estonian": "est",
    "Basque": "eus",
    "Persian": "fas",
    "Finnish": "fin",
    "French": "fra",
    "Western Frisian": "fry",
    "Irish": "gle",
    "Scottish Gaelic": "gla",
    "Galician": "glg",
    "Gujarati": "guj",
    "Hausa": "hau",
    "Hebrew": "heb",
    "Hindi": "hin",
    "Croatian": "hrv",
    "Hungarian": "hun",
    "Armenian": "hye",
    "Indonesian": "ind",
    "Icelandic": "isl",
    "Italian": "ita",
    "Japanese": "jpn",
    "Javanese": "jav",
    "Georgian": "kat",
    "Kazakh": "kaz",
    "Khmer": "khm",
    "Kannada": "kan",
    "Korean": "kor",
    "Kurdish": "kur",
    "Kyrgyz": "kir",
    "Latin": "lat",
    "Lao": "lao",
    "Lithuanian": "lit",
    "Latvian": "lav",
    "Malagasy": "mlg",
    "Macedonian": "mkd",
    "Malayalam": "mal",
    "Mongolian": "mon",
    "Marathi": "mar",
    "Malay": "msa",
    "Burmese": "mya",
    "Nepali": "nep",
    "Dutch": "nld",
    "Norwegian": "nor",
    "Oromo": "orm",
    "Oriya": "ori",
    "Punjabi": "pan",
    "Polish": "pol",
    "Pashto": "pus",
    "Portuguese": "por",
    "Romanian": "ron",
    "Russian": "rus",
    "Sanskrit": "san",
    "Sindhi": "snd",
    "Sinhala": "sin",
    "Slovak": "slk",
    "Slovenian": "slv",
    "Somali": "som",
    "Albanian": "sqi",
    "Serbian": "srp",
    "Sundanese": "sun",
    "Swedish": "swe",
    "Swahili": "swa",
    "Tamil": "tam",
    "Telugu": "tel",
    "Thai": "tha",
    "Tagalog": "tgl",
    "Turkish": "tur",
    "Uyghur": "uig",
    "Ukrainian": "ukr",
    "Urdu": "urd",
    "Uzbek": "uzb",
    "Vietnamese": "vie",
    "Xhosa": "xho",
    "Yiddish": "yid",
    "Chinese": "chi_sim",
}

TESSERACT_CODE_TO_LANGUAGE = {v: k for k, v in LANGUAGE_TO_TESSERACT_CODE.items()}

CODE_TO_LANGUAGE = {
    "_math": "Math",
    "af": "Afrikaans",
    "am": "Amharic",
    "ar": "Arabic",
    "as": "Assamese",
    "az": "Azerbaijani",
    "be": "Belarusian",
    "bg": "Bulgarian",
    "bn": "Bengali",
    "br": "Breton",
    "bs": "Bosnian",
    "ca": "Catalan",
    "cs": "Czech",
    "cy": "Welsh",
    "da": "Danish",
    "de": "German",
    "el": "Greek",
    "en": "English",
    "eo": "Esperanto",
    "es": "Spanish",
    "et": "Estonian",
    "eu": "Basque",
    "fa": "Persian",
    "fi": "Finnish",
    "fr": "French",
    "fy": "Western Frisian",
    "ga": "Irish",
    "gd": "Scottish Gaelic",
    "gl": "Galician",
    "gu": "Gujarati",
    "ha": "Hausa",
    "he": "Hebrew",
    "hi": "Hindi",
    "hr": "Croatian",
    "hu": "Hungarian",
    "hy": "Armenian",
    "id": "Indonesian",
    "is": "Icelandic",
    "it": "Italian",
    "ja": "Japanese",
    "jv": "Javanese",
    "ka": "Georgian",
    "kk": "Kazakh",
    "km": "Khmer",
    "kn": "Kannada",
    "ko": "Korean",
    "ku": "Kurdish",
    "ky": "Kyrgyz",
    "la": "Latin",
    "lo": "Lao",
    "lt": "Lithuanian",
    "lv": "Latvian",
    "mg": "Malagasy",
    "mk": "Macedonian",
    "ml": "Malayalam",
    "mn": "Mongolian",
    "mr": "Marathi",
    "ms": "Malay",
    "my": "Burmese",
    "ne": "Nepali",
    "nl": "Dutch",
    "no": "Norwegian",
    "om": "Oromo",
    "or": "Oriya",
    "pa": "Punjabi",
    "pl": "Polish",
    "ps": "Pashto",
    "pt": "Portuguese",
    "ro": "Romanian",
    "ru": "Russian",
    "sa": "Sanskrit",
    "sd": "Sindhi",
    "si": "Sinhala",
    "sk": "Slovak",
    "sl": "Slovenian",
    "so": "Somali",
    "sq": "Albanian",
    "sr": "Serbian",
    "su": "Sundanese",
    "sv": "Swedish",
    "sw": "Swahili",
    "ta": "Tamil",
    "te": "Telugu",
    "th": "Thai",
    "tl": "Tagalog",
    "tr": "Turkish",
    "ug": "Uyghur",
    "uk": "Ukrainian",
    "ur": "Urdu",
    "uz": "Uzbek",
    "vi": "Vietnamese",
    "xh": "Xhosa",
    "yi": "Yiddish",
    "zh": "Chinese",
}

LANGUAGE_TO_CODE = {v: k for k, v in CODE_TO_LANGUAGE.items()}



from types import MappingProxyType
from typing import Any

import numpy as np
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.params import CODE_TO_LANGUAGE, LANGUAGE_TO_TESSERACT_CODE
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document

from ...enums import OEM, PSM, TessLib
from .BaseRecognizer import BaseRecognizer


class TesseractRecognizer(BaseRecognizer):
    """
    Run Tesseract text recognition on images.
    """

    oem = Param(
        Params._dummy(),
        "oem",
        "OCR engine mode. Defaults to :attr:`OEM.DEFAULT`.",
        typeConverter=TypeConverters.toInt,
    )

    tessDataPath = Param(
        Params._dummy(),
        "tessDataPath",
        "Path to tesseract data folder.",
        typeConverter=TypeConverters.toString,
    )

    tessLib = Param(
        Params._dummy(),
        "tessLib",
        "The desired Tesseract library to use. Defaults to :attr:`TESSEROCR`",
        typeConverter=TypeConverters.toInt,
    )

    defaultParams = MappingProxyType(
        {
            "inputCols": ["image", "boxes"],
            "outputCol": "text",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "oem": OEM.DEFAULT,
            "lang": ["eng"],
            "lineTolerance": 0,
            "keepFormatting": False,
            "tessDataPath": "/usr/share/tesseract-ocr/5/tessdata/",
            "tessLib": TessLib.PYTESSERACT,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(TesseractRecognizer, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_pytesseract(cls, images, boxes, params):
        raise NotImplementedError("Pytesseract version not implemented yet.")

    @staticmethod
    def getLangTess(params):
        return "+".join(
            [
                LANGUAGE_TO_TESSERACT_CODE[CODE_TO_LANGUAGE[lang]]
                for lang in params["lang"]
            ],
        )

    @classmethod
    def call_pytesseract(cls, images, detected_boxes, params):  # pragma: no cover
        import cv2
        from tesserocr import PyTessBaseAPI

        results = []
        box_id = 0

        for (image, image_path), detected_box in zip(images, detected_boxes):
            # api.SetImage(image)

            boxes = []
            texts = []

            image_np = np.array(image)

            for b in detected_box.bboxes:
                box = b
                if isinstance(box, dict):
                    box = Box(**box)
                if not isinstance(box, Box):
                    box = Box(**box.asDict())

                # Scale the bounding box first (if params["scaleFactor"] is used)
                # Note: 'padding' in box.scale might add border, consider its effect on angle if it's applied after rotation
                # For simplicity, assuming scaled_box gives the final desired dimensions and position.
                scaled_box = box.scale(params["scaleFactor"], padding=0)

                center_x = scaled_box.x + scaled_box.width / 2
                center_y = scaled_box.y + scaled_box.height / 2
                center_tuple = (float(center_x), float(center_y))

                size_tuple = (float(scaled_box.width), float(scaled_box.height))


                if scaled_box.angle == 90:
                    scaled_box.angle = -90

                rect = (center_tuple, size_tuple, scaled_box.angle)

                src_pts = cv2.boxPoints(rect).astype("float32")

                # Define the destination points for the output cropped image
                # This will be an upright rectangle with dimensions (width x height)
                output_width = int(scaled_box.width)
                output_height = int(scaled_box.height)

                # # dst_pts order: Bottom-Left, Top-Left, Top-Right, Bottom-Right to match cv2.boxPoints
                dst_pts = np.array(
                    [
                        [0, output_height - 1],
                        [0, 0],
                        [output_width - 1, 0],
                        [output_width - 1, output_height - 1],
                    ],
                    dtype="float32",
                )
                try:
                    M_transform = cv2.getPerspectiveTransform(src_pts, dst_pts)
                except cv2.error as e:
                    print(
                        f"Error calculating perspective transform for box {box}: {e}. Skipping.",
                    )
                    continue

                try:
                    cropped_rotated_as_is_np = cv2.warpPerspective(
                        image_np,
                        M_transform,
                        (output_width, output_height),
                        flags=cv2.INTER_CUBIC,
                        borderMode=cv2.BORDER_REPLICATE,
                    )
                except cv2.error as e:
                    print(
                        f"Error during perspective warp for box {box}: {e}. Skipping.",
                    )
                    continue

                if (
                    cropped_rotated_as_is_np.shape[0] == 0
                    or cropped_rotated_as_is_np.shape[1] == 0
                ):
                    print(
                        f"Warning: Cropped image 'as is' for box {box} is empty. Skipping.",
                    )
                    continue

                final_image_for_ocr_np = None
                if cropped_rotated_as_is_np.ndim == 3:
                    if cropped_rotated_as_is_np.shape[2] == 4:
                        final_image_for_ocr_np = cv2.cvtColor(
                            cropped_rotated_as_is_np,
                            cv2.COLOR_RGBA2RGB,
                        )
                    elif cropped_rotated_as_is_np.shape[2] == 3:
                        final_image_for_ocr_np = cropped_rotated_as_is_np
                    else:
                        print(
                            f"Warning: Unexpected channels {cropped_rotated_as_is_np.shape[2]}. Converted to grayscale.",
                        )
                        final_image_for_ocr_np = cv2.cvtColor(
                            cropped_rotated_as_is_np,
                            cv2.COLOR_BGR2GRAY,
                        )
                elif cropped_rotated_as_is_np.ndim == 2:
                    final_image_for_ocr_np = cropped_rotated_as_is_np
                else:
                    print(
                        f"Error: Cropped image has unexpected dimensions: {cropped_rotated_as_is_np.shape}. Cannot process.",
                    )
                    continue
                # --- Prepare the "as is" cropped image for OCR API ---
                # Tesseract/Leptonica often prefers 8-bit grayscale or 24-bit RGB.
                # Still need to handle potential format issues like 32-bit RGBA.

                # Now, send this cropped_rotated_image to your OCR API
                # The specific method will depend on your 'api' object.
                # Assuming your 'api' has a method to set an image directly (like a PIL Image or NumPy array)
                # You might need to convert `cropped_rotated_image` to a compatible format.

                # Example for Tesseract (pytesseract):
                # If your `api` is a pytesseract-like wrapper, it might have an `image_to_string` method
                import pytesseract
                from PIL import Image

                pil_image = Image.fromarray(final_image_for_ocr_np)


                data = pytesseract.image_to_data(
                    pil_image,
                    lang=cls.getLangTess(params),
                    config="--psm 8",
                    output_type=pytesseract.Output.DICT,
                )
                # Get text and confidence from the first text element (if exists)

                if len(data["text"]) > 0:
                    text_entries = []
                    scores = []
                    for i in range(len(data["text"])):
                        if data["level"][i] == 5 and data["text"][i].strip():
                            text_entries.append(data["text"][i].replace("\n", ""))
                            scores.append(
                                (
                                    float(data["conf"][i]) / 100
                                    if data["conf"][i] != "-1"
                                    else 0
                                ),
                            )

                    if text_entries:
                        box.text = " ".join(text_entries)
                        box.score = sum(scores) / len(
                            scores,
                        )  # Average confidence score
                else:
                    box.text = ""
                    box.score = 0

                if box.score > params["scoreThreshold"]:
                    boxes.append(box)
                    texts.append(box.text)

            if params["keepFormatting"]:
                text = TesseractRecognizer.box_to_formatted_text(
                    boxes,
                    params["lineTolerance"],
                )
            else:
                text = " ".join(texts)

            results.append(
                Document(
                    path=image_path,
                    text=text,
                    bboxes=boxes,
                    type="text",
                    exception="",
                ),
            )
        return results

    @classmethod
    def call_tesserocr(cls, images, detected_boxes, params):  # pragma: no cover
        from tesserocr import PyTessBaseAPI

        results = []
        lang = cls.getLangTess(params)
        with PyTessBaseAPI(
            path=params["tessDataPath"],
            psm=PSM.SINGLE_WORD,
            oem=params["oem"],
            lang=lang,
        ) as api:
            api.SetVariable("debug_file", "ocr.log")

            for (image, image_path), detected_box in zip(images, detected_boxes):
                api.SetImage(image)

                boxes = []
                texts = []

                for b in detected_box.bboxes:
                    box = b
                    if isinstance(box, dict):
                        box = Box(**box)
                    if not isinstance(box, Box):
                        box = Box(**box.asDict())
                    scaled_box = box.scale(params["scaleFactor"], padding=0)
                    api.SetRectangle(
                        scaled_box.x,
                        scaled_box.y,
                        scaled_box.width,
                        scaled_box.height,
                    )
                    box.text = api.GetUTF8Text().replace("\n", "")
                    box.score = api.MeanTextConf() / 100
                    if box.score > params["scoreThreshold"]:
                        boxes.append(box)
                        texts.append(box.text)
                if params["keepFormatting"]:
                    text = TesseractRecognizer.box_to_formatted_text(
                        boxes,
                        params["lineTolerance"],
                    )
                else:
                    text = " ".join(texts)

                results.append(
                    Document(
                        path=image_path,
                        text=text,
                        bboxes=boxes,
                        type="text",
                        exception="",
                    ),
                )
        return results

    @classmethod
    def call_recognizer(cls, images, boxes, params):
        if params["tessLib"] == TessLib.TESSEROCR.value:
            return cls.call_tesserocr(images, boxes, params)
        if params["tessLib"] == TessLib.PYTESSERACT.value:
            return cls.call_pytesseract(images, boxes, params)
        raise ValueError(f"Unknown Tesseract library: {params['tessLib']}")

    def setOem(self, value):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self._set(oem=value)

    def getOem(self):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self.getOrDefault(self.oem)

    def setTessDataPath(self, value):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self._set(tessDataPath=value)

    def getTessDataPath(self):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self.getOrDefault(self.tessDataPath)

    def setTessLib(self, value):
        """
        Sets the value of :py:attr:`tessLib`.
        """
        return self._set(tessLib=value)

    def getTessLib(self):
        """
        Gets the value of :py:attr:`tessLib`.
        """
        return self.getOrDefault(self.tessLib)



from typing import Any, Dict, List

from Levenshtein import distance
from pydantic import BaseModel


def calculate_similarity(obj1: BaseModel, obj2: BaseModel) -> float:
    """
    Calculates the similarity between two instances of a Pydantic model.

    Args:
      obj1: The first Pydantic model instance.
      obj2: The second Pydantic model instance.

    Returns:
      A float value representing the similarity between the two objects.
      Higher values indicate greater similarity.

    """

    # 1. Get dictionaries of model attributes
    obj1_dict = obj1.dict()
    obj2_dict = obj2.dict()

    # 2. Find common attributes
    common_attributes = set(obj1_dict.keys()).intersection(set(obj2_dict.keys()))

    # 3. Calculate similarity based on common attributes
    similarity_score = 0
    for attr in common_attributes:
        value1 = obj1_dict[attr]
        value2 = obj2_dict[attr]

        similarity_score += _calculate_value_similarity(value1, value2)

    # 4. Normalize similarity score
    similarity_score /= len(common_attributes)

    return similarity_score


def _calculate_value_similarity(value1: Any, value2: Any) -> float:
    """
    Calculates the similarity between two values.

    Handles different data types, including nested objects within lists and dictionaries.
    """
    if isinstance(value1, BaseModel):
        return calculate_similarity(value1, value2)
    if isinstance(value1, list):
        return _calculate_list_similarity(value1, value2)
    if isinstance(value1, dict):
        return _calculate_dict_similarity(value1, value2)
    if isinstance(value1, (int, float)):
        return 1 - abs(value1 - value2) / max(abs(value1), abs(value2), 1)
    if isinstance(value1, str):
        if value1 == "" and value2 == "":
            return 1.0
        return 1 - distance(value1, value2) / max(len(value1), len(value2))
    # Handle other data types (e.g., booleans, dates) as needed
    return int(value1 == value2)


def _calculate_list_similarity(list1: List[Any], list2: List[Any]) -> float:
    """
    Calculates the similarity between two lists.

    Handles nested objects within lists.
    """
    if not list1 and not list2:
        return 1.0  # Both lists are empty
    if not list1 or not list2:
        return 0.0  # One list is empty

    min_len = min(len(list1), len(list2))
    max_len = max(len(list1), len(list2))

    total_similarity = 0
    for i in range(min_len):
        total_similarity += _calculate_value_similarity(list1[i], list2[i])

    # Handle cases where lists have different lengths
    total_similarity /= max_len

    return total_similarity


def _calculate_dict_similarity(dict1: Dict[Any, Any], dict2: Dict[Any, Any]) -> float:
    """
    Calculates the similarity between two dictionaries.

    Handles nested objects within dictionaries.
    """
    common_keys = set(dict1.keys()).intersection(set(dict2.keys()))
    if not common_keys:
        return 0

    total_similarity = 0
    for key in common_keys:
        total_similarity += _calculate_value_similarity(dict1[key], dict2[key])

    return total_similarity / len(common_keys)



import json
import logging
import traceback
from typing import Optional

from pydantic import BaseModel
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import lit, udf
from sparkdantic import create_spark_schema

from scaledp.params import (
    HasColumnValidator,
    HasDefaultEnum,
    HasInputCol,
    HasKeepInputData,
    HasNumPartitions,
    HasOutputCol,
    HasPageCol,
    HasPathCol,
    HasPropagateExc,
    HasSchema,
)
from scaledp.schemas.Image import Image


class VisualExtractorError(Exception):
    pass


class BaseVisualExtractor(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasPathCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasNumPartitions,
    HasPageCol,
    HasColumnValidator,
    HasDefaultEnum,
    HasPropagateExc,
    HasSchema,
):

    def get_params(self):
        return json.dumps({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def call_extractor(cls, images, params):
        raise NotImplementedError("Subclasses should implement this method")

    def transform_udf(self, image, params=None):
        logging.info("Run Image Data Extractor")
        if params is None:
            params = self.get_params()
        params = json.loads(params)
        if not isinstance(image, Image):
            image = Image(**image.asDict())
        if image.exception != "":
            return self.getOutputClass()(
                path=image.path,
                type="extractor",
                exception=image.exception,
            )
        try:

            result = self.call_extractor([image], params)
        except Exception as e:
            exception = traceback.format_exc()
            exception = (
                f"{self.uid}: Error in data extraction: {exception}, {image.exception}"
            )
            logging.warning(f"{self.uid}: Error in data extraction.")
            if self.getPropagateError():
                raise VisualExtractorError from e
            return self.getOutputClass()(
                path=image.path,
                type="detector",
                exception=exception,
            )
        return result[0]

    def getOutputClass(self):
        class ExtractorOutputCustom(BaseModel):
            path: str
            json_data: Optional[str] = ""
            type: str
            exception: str = ""
            processing_time: float = 0.0
            data: Optional[self.getPaydanticSchema()] = None

        return ExtractorOutputCustom

    def get_output_schema(self):
        return create_spark_schema(self.getOutputClass())

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        in_col = self._validate(self.getInputCol(), dataset)

        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, self.get_output_schema())(
                in_col,
                lit(self.get_params()),
            ),
        )
        if not self.getKeepInputData():
            result = result.drop(in_col)
        return result



from datetime import date, datetime, time
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel, Field, create_model


def json_schema_to_model(
    json_schema: Dict[str, Any],
    definitions: Optional[Dict[str, Any]] = None,
) -> Type[BaseModel]:
    """
    Converts a JSON schema to a Pydantic BaseModel class.

    Args:
        json_schema: The JSON schema to convert.
        definitions: A dictionary of defined schemas, often found in the "$defs" property.

    Returns:
        A Pydantic BaseModel class.

    """

    model_name = json_schema.get("title")
    field_definitions = {
        name: json_schema_to_pydantic_field(
            name,
            prop,
            json_schema.get("required", []),
            definitions,
        )
        for name, prop in json_schema.get("properties", {}).items()
    }
    return create_model(model_name, **field_definitions)


def json_schema_to_pydantic_field(
    name: str,
    json_schema: Dict[str, Any],
    required: List[str],
    definitions: Optional[Dict[str, Any]] = None,
) -> Any:
    """
    Converts a JSON schema property to a Pydantic field definition.

    Args:
        name: The field name.
        json_schema: The JSON schema property.
        required: A list of required field names.
        definitions: A dictionary of defined schemas, often found in the "$defs" property.

    Returns:
        A Pydantic field definition.

    """

    type_ = json_schema_to_pydantic_type(json_schema, definitions)
    description = json_schema.get("description")
    examples = json_schema.get("examples")
    min_length = json_schema.get("minLength")
    max_length = json_schema.get("maxLength")
    minimum = json_schema.get("minimum")
    maximum = json_schema.get("maximum")
    exclusive_minimum = json_schema.get("exclusiveMinimum")
    exclusive_maximum = json_schema.get("exclusiveMaximum")
    pattern = json_schema.get("pattern")

    # Handle optional fields (including 'anyOf' with 'null')
    if name not in required or (
        "anyOf" in json_schema
        and any(item.get("type") == "null" for item in json_schema.get("anyOf", []))
    ):
        type_ = Optional[type_]

    return (
        type_,
        Field(
            description=description,
            examples=examples,
            min_length=min_length,
            max_length=max_length,
            ge=minimum,
            le=maximum,
            gt=exclusive_minimum,
            lt=exclusive_maximum,
            pattern=pattern,
            default=... if name in required else None,
        ),
    )


def json_schema_to_pydantic_type(
    json_schema: Dict[str, Any],
    definitions: Optional[Dict[str, Any]] = None,
) -> Any:
    """
    Converts a JSON schema type to a Pydantic type.

    Args:
        json_schema: The JSON schema to convert.
        definitions: A dictionary of defined schemas, often found in the "$defs" property.

    Returns:
        A Pydantic type.

    """

    type_ = json_schema.get("type")

    if type_ == "string":
        format_ = json_schema.get("format")
        if format_ == "date":
            return date
        if format_ == "time":
            return time
        if format_ == "date-time":
            return datetime
        if "enum" in json_schema:
            return Enum(
                json_schema.get("title", "EnumType"),
                {member: member for member in json_schema["enum"]},
            )
        return str
    if type_ == "integer":
        return int
    if type_ == "number":
        return float
    if type_ == "boolean":
        return bool
    if type_ == "array":
        items_schema = json_schema.get("items")
        if items_schema:
            item_type = json_schema_to_pydantic_type(items_schema, definitions)
            return List[item_type]
        return List
    if type_ == "object":
        # Handle nested models.
        properties = json_schema.get("properties")
        if properties:
            return json_schema_to_model(json_schema, definitions)
        return Dict
    if type_ == "null":
        return None
    if "$ref" in json_schema:
        # Handle references to nested schemas
        ref_path = json_schema["$ref"].split("/")
        ref_name = ref_path[-1]
        if definitions:
            ref_schema = definitions.get(ref_name)
            if ref_schema:
                return json_schema_to_pydantic_type(ref_schema, definitions)
        raise ValueError(f"Could not resolve reference: {json_schema['$ref']}")
    if "anyOf" in json_schema:
        # Handle 'anyOf' with 'null' for optional fields
        types = [
            json_schema_to_pydantic_type(item, definitions)
            for item in json_schema["anyOf"]
        ]
        if any(t is None for t in types):  # If 'null' is present in anyOf
            types = [t for t in types if t is not None]  # Remove 'null' from the list
            if len(types) == 1:
                return Optional[types[0]]  # If only one non-null type, use Optional
            return Union[types]  # If multiple non-null types, use Union
        # If 'null' is not present, return the union of all types
        return Union[types]
    raise ValueError(f"Unsupported JSON schema type: {type_}")
